
/*  @(#)run.c 1.7 91/09/23
 *
 *  Run time interpreter used by the popi program.
 *
 *  Popi was originally written by Gerard J. Holzmann - AT&T Bell Labs.
 *  This version is based on the code in his Prentice Hall book,
 *  "Beyond Photography - the digital darkroom," ISBN 0-13-074410-7,
 *  which is copyright (c) 1988 by Bell Telephone Laboratories, Inc. 
 *
 *  Permission is given to distribute these extensions, as long as these
 *  introductory messages are not removed, and no monies are exchanged.
 *
 *  No responsibility is taken for any errors or inaccuracies inherent
 *  either to the comments or the code of this program, but if reported
 *  (see README file) then an attempt will be made to fix them.
 */

#include <ctype.h>
#include "popi.h"

#if        SEQPAR
#include <parallel/microtask.h>
#endif  /* SEQPAR */

#define  STACKSIZE  128

#define  dop(OP)    a = *--sp ; tp = sp-1 ; *tp = (*tp OP (stack_t) a)
#if defined (ISTACK)
#define  diop(OP)   dop(OP)
#else
#define  diop(OP)   ca = *--sp ; tp = sp-1 ; cb = *tp ; *tp = (cb OP ca)
#endif /*ISTACK*/

/* Local function prototypes */

static void prun P((void)) ;


void
SwapOldNew()
{
  struct SRC *tmp ;

  tmp = CurOld ;
  CurOld = CurNew ;
  CurNew = tmp ;
  STRCPY(CurOld->str, "old") ;
  STRCPY(CurNew->str, "new") ;
}


static void
prun()
{
  long *ParseEnd ;     /* After end of parse string */
  int color ;
  int x, y ;           /* Coordinates */
  int nrange = 0 ;     /* No. range errors */

  ParseEnd = &parsed[prs] ;

#ifndef NDEBUG
  if (Debug)
    {
      register long *CurrParse ;  /* Pointer to current item in parse string */

      FPRINTF(Debug, "Parse string\n") ;
      for (CurrParse = parsed; CurrParse != ParseEnd; ++CurrParse)
        FPRINTF(Debug, "'%c' (%ld)\n", (char) *CurrParse, *CurrParse) ;
      FPRINTF(Debug, "---\n") ;
    }
#endif /* NDEBUG */
  DEBUG((Debug, "computing %i colors\n", col_used)) ;

/*  Warning: Microsoft Quick C generates incorrect code for the following
 *  loop when optimisation is turned on.
 */

#ifndef SEQPAR
    for (y = 0; y < CurNew->height; ++y)
#else  /* ! SEQPAR */
    for (y = m_myid; y < CurNew->height; y += m_numprocs)
#endif  /* ! SEQPAR */
      {
#if defined(ISTACK)
        short *ap ;    /* Precalculated polar angles */
        short *rp ;    /* Precalculated polar radius */
#else
	float *ap ;
	float *rp ;
#endif /*ISTACK*/
        register pixel_t *pixp ;            /* Default destination. */
	register pixel_t *pixaddr ;         /* Explicit destination. */
	register union IntFloat intfloat ;  /* Get floats from array of long. */

        ap = &avals[y * CurNew->width] ;
        rp = &rvals[y * CurNew->width] ;

        for (color = 0; color < col_used; ++color)
          {
            pixp = &CurNew->pix[color][y*Xsize] ;

            for (x = 0; x < CurNew->width; ++x, ++pixp)
              {
                stack_t Stack[STACKSIZE] ;   /* The stack */
                register stack_t *sp ;       /* Stack pointer (top of stack) */
                register stack_t *tp ;       /* Temp stack pointer */
                stack_t a ;
                stack_t b ;
                int c ;                      /* Scratch */
                long ca, cb ;
                register long *CurrParse ;   /* Pointer to current item */
                                             /* in parse string */
		double dt ;		     /* overflow tester */

                for (CurrParse = parsed, sp = Stack; CurrParse != ParseEnd;
                     ++CurrParse)
                  {
                    switch ((int) *CurrParse)
                      {

        case OP_VALUE : *sp++ = (stack_t) *++CurrParse ;
                        continue ;

	case OP_VALF  : intfloat.intgr = *++CurrParse ;
                        *sp++ = (stack_t) intfloat.flt ;
                        continue ;

        case OP_AT    : --sp ;
                        if (Truncate)
                          {
                                 if (*sp > (stack_t) Zmax) *sp = Zmax ;
                            else if (*sp < 0) *sp = 0 ;
                          }
                        *pixp = (pixel_t) *sp ;
                        continue ;

        case OP_PLUS  : dop(+) ;
                        break ;

        case OP_MINUS : dop(-) ;
                        break ;

        case OP_MUL   : dop(*) ;
                        break ;

        case OP_DIV   : a = *--sp ;
                        tp = sp-1 ;
                        if (a == 0) *tp = Zmax ;
                        else *tp = (*tp / (stack_t) a) ;
                        break ;

        case OP_MOD   : ca = *--sp ;
                        tp = sp-1 ;
                        if (ca == 0) *tp = 0 ;
                        else
                          {
			    cb = *tp ;
			    *tp = (cb % ca) ;
                          }
                        break ;

        case OP_GT    : dop(>) ;
                        break ;

        case OP_LT    : dop(<) ;
                        break ;

        case OP_GE    : dop(>=) ;
                        break ;

        case OP_LE    : dop(<=) ;
                        break ;

        case OP_EQ    : dop(==) ;
                        break ;

        case OP_NE    : dop(!=) ;
                        break ;

        case OP_CAND  : diop(&&) ;
                        break ;

        case OP_COR   : diop(||) ;
                        break ;

        case OP_XOR   : diop(^) ;
                        break ;

        case OP_OR    : diop(|) ;
                        break ;

        case OP_AND   : diop(&) ;
                        break ;

        case OP_X     : *sp++ = (stack_t) x ;
                        break ;

        case OP_Y     : *sp++ = (stack_t) y ;
                        break ;

        case OP_UMIN  : tp = sp-1 ;
                        *tp = -(*tp) ;
                        break ;

        case OP_NOT   : tp = sp-1 ;
                        *tp = !(*tp) ;
                        break ;

        case OP_ASGN  : a = *--sp ;
			--sp ;
                        if (Truncate)
                          {
                                 if (a > (stack_t) Zmax) a = Zmax ;
                            else if (a < 0) a = 0 ;
                          }
                        if (pixaddr) *pixaddr = (pixel_t) a ;
                        break ;

        case OP_CRVAL : ca = *--sp ;              /* y */
                        cb = *--sp ;              /* x */
                        tp = sp-1 ;
                        c = (int) *tp ;
			if (ca >= Images[c].height  || ca < 0 ||
                            cb >= Images[c].width   || cb < 0)
			  {
			    if (RangeCheck && (nrange++ == 0))
			      FPRINTF(stderr,
                                      "Range err at (%d,%d) => %s[%d, %d]\n",
                                      x, y, Images[c].str, cb, ca) ;
		            switch (rng_policy)
			      {
			        case CUT :
				  *tp = (stack_t) 0 ;
				  break ;
				  
				case WRAP :
				  if (ca >= Images[c].height) ca = ca % Images[c].height ;
				  if (ca < 0) ca = Images[c].height + (++ca % Images[c].height) - 1 ;
				  if (cb >= Images[c].width)  cb = cb % Images[c].width ;
				  if (cb < 0) cb = Images[c].width + (++cb % Images[c].width) - 1 ;
				  *tp = (stack_t) Images[c].pix[color][(ca*Images[c].Xalloc)+cb] ;
				  break;
				  
				case MINMAX :
				  if (ca >= Images[c].height) ca = Images[c].height - 1 ;
                                  if (ca < 0)                 ca = 0 ;
                                  if (cb >= Images[c].width)  cb = Images[c].width - 1 ;
                                  if (cb < 0)                 cb = 0 ;
				  *tp = (stack_t) Images[c].pix[color][(ca*Images[c].Xalloc)+cb] ;
				  break ;
			      }
			  }
			else
			  *tp = (stack_t) Images[c].pix[color][(ca*Images[c].Xalloc)+cb] ;
                        break ;

        case OP_CLVAL : ca = *--sp ;              /* y */
                        cb = *--sp ;              /* x */
                        tp = sp-1 ;
                        c = (int) *tp ;
			if (ca >= Images[c].height  || ca < 0 ||
                            cb >= Images[c].width   || cb < 0)
			  {
			    if (RangeCheck && (nrange++ == 0))
			      FPRINTF(stderr,
                                      "Range err at (%d,%d) => %s[%d, %d]\n",
                                      x, y, Images[c].str, cb, ca) ;
		            switch (rng_policy)
			      {
			        case CUT :
				case MINMAX :
				  pixaddr = NULL ;
				  break ;
				  
				case WRAP :
				  if (ca >= Images[c].height) ca = ca % Images[c].height ;
				  if (ca < 0) ca = Images[c].height + (++ca % Images[c].height) - 1 ;
				  if (cb >= Images[c].width)  cb = cb % Images[c].width ;
				  if (cb < 0) cb = Images[c].width + (++cb % Images[c].width) - 1 ;
                                  pixaddr = &(Images[c].pix[color][(ca*Images[c].Xalloc)+cb]) ;
				  break;
			      }
			  }
			else
                          pixaddr = &(Images[c].pix[color][(ca*Images[c].Xalloc)+cb]) ;
                        break ;

        case OP_PRVAL : {
                          int xval ;
                          int yval ;
    
                          a = *--sp ;    /* Angle */
                          b = *--sp ;    /* Radius */

                          tp = sp - 1 ;
                          c = (int) *tp ;

                          xval = (int) (b * cos((double) DtoR(a)) +
                                        Images[c].width / 2.0) ;
                          yval = (int) (- b * sin((double) DtoR(a)) +
                                        Images[c].height / 2.0) ;

			if (yval >= Images[c].height  || yval < 0 ||
                            xval >= Images[c].width   || xval < 0)
			  {
			    if (RangeCheck && (nrange++ == 0))
                              FPRINTF(stderr,
                                 "Range err at (%d,%d) => %s{%d,%d} [%d,%d]\n",
                                 x, y, Images[c].str, b, a, xval, yval) ;
		            switch (rng_policy)
			      {
			        case CUT :
				  *tp = (stack_t) 0 ;
				  break ;
				  
				case WRAP :
				  if (yval >= Images[c].height) yval = yval % Images[c].height ;
				  if (yval < 0) yval = Images[c].height + (++yval % Images[c].height) - 1 ;
				  if (xval >= Images[c].width)  xval = xval % Images[c].width ;
				  if (xval < 0) xval = Images[c].width + (++xval % Images[c].width) - 1 ;
				  *tp = (stack_t) Images[c].pix[color][(yval*Images[c].Xalloc)+xval] ;
				  break;
				  
				case MINMAX :
				  if (yval >= Images[c].height) yval = Images[c].height - 1 ;
                                  if (yval < 0)                 yval = 0 ;
                                  if (xval >= Images[c].width)  xval = Images[c].width - 1 ;
                                  if (xval < 0)                 xval = 0 ;
				  *tp = (stack_t) Images[c].pix[color][(yval*Images[c].Xalloc)+xval] ;
				  break ;
			      }
			  }
			else
			  *tp = (stack_t) Images[c].pix[color][(yval*Images[c].Xalloc)+xval] ;
			  }
                        break ;

        case OP_PLVAL : {
                          int xval ;
                          int yval ;

                          a = *--sp ;    /* Angle */
                          b = *--sp ;    /* Radius */

                          tp = sp - 1 ;
                          c = (int) *tp ;

                          xval = (int) (b * cos((double) DtoR(a)) +
                                        Images[c].width / 2.0) ;
                          yval = (int) (- b * sin((double) DtoR(a)) +
                                        Images[c].height / 2.0) ;

			if (yval >= Images[c].height  || yval < 0 ||
                            xval >= Images[c].width   || xval < 0)
			  {
			    if (RangeCheck && (nrange++ == 0))
                              FPRINTF(stderr,
                                  "Range err at (%d,%d) => %s{%d,%d} [%d,%d]\n",
                                  x, y, Images[c].str, b, a, xval, yval) ;
		            switch (rng_policy)
			      {
			        case CUT :
				case MINMAX :
				  pixaddr = NULL ;
				  break ;
				  
				case WRAP :
				  if (yval >= Images[c].height) yval = yval % Images[c].height ;
				  if (yval < 0) yval = Images[c].height + (++yval % Images[c].height) - 1 ;
				  if (xval >= Images[c].width)  xval = xval % Images[c].width ;
				  if (xval < 0) xval = Images[c].width + (++xval % Images[c].width) - 1 ;
                                  pixaddr = &(Images[c].pix[color][(yval*Images[c].Xalloc)+xval]) ;
				  break;
			      }
			  }
			else
                          pixaddr = &(Images[c].pix[color][(yval*Images[c].Xalloc)+xval]) ;
			  }
                        break ;

        case OP_POW   : a = *--sp ;     /* Exponent */
                        tp = sp-1 ;
			/* test for conversion overflow */
                        if ( *tp != 0 &&
			     (dt = pow((double) *tp, (double) a)) <=
				(double)MAXLONG &&
			     dt >= -MAXLONG )
			  {
			    *tp = (stack_t) dt ;
			  }	/* else leave *tp as is? */
                        break ;

        case OP_A     :
#if        NOMEM
                        *sp++ = (stack_t)
                                RtoD(atan2((double) y, (double) x)) + 0.5 ;
#else  /* !NOMEM */
                        *sp++ = (stack_t) *ap ;
#endif  /* !NOMEM */
                        break ;

        case OP_R     : *sp++ = (stack_t) *rp ;
                        break ;

        case OP_SIN   : tp = sp-1 ;
                        *tp = (stack_t)
                              (sin((double) DtoR(*tp))
#if defined(ISTACK)	/* on integer stack scale it by Z */
					* (double) Zmax
#endif
							) ;
                        break ;

        case OP_COS   : tp = sp-1 ;
                        *tp = (stack_t)
                              (cos((double) DtoR(*tp))
#if defined(ISTACK)     /* on integer stack scale it by Z */
					* (double) Zmax
#endif
							) ;
                        break ;

        case OP_ATAN  : a = *--sp ;
                        tp = sp-1 ;
                        *tp = (stack_t)
                              RtoD(atan2((double) *tp, (double) a)) ;
                        break ;

        case OP_HYPOT : a = *--sp ;
                        tp = sp - 1 ;
                        *tp = (stack_t) hypot((double) *tp, (double) a) ;
                        break ;

        case OP_ABS   : tp = sp-1 ;
                        *tp = *tp < 0 ? - *tp : *tp ;
                        break ;

        case OP_LOG   : tp = sp-1 ;
			if ( *tp > 0 )
			    *tp = (stack_t) log((double) *tp) ;
                        break ;

        case OP_SQRT  : tp = sp-1 ;
			if ( *tp > 0 )
			    *tp = (stack_t) sqrt((double) *tp) ;
                        break ;

        case OP_RAND  : *sp++ = (stack_t) RANDOM ;
                        break ;

        case OP_LSHIFT : diop(<<) ;
                         break ;

        case OP_RSHIFT : diop(>>) ;
                         break ;

        case OP_CONDIT : ca = *--sp ;
                         CurrParse++ ;
                         if (!ca)
                           CurrParse = &parsed[*CurrParse] ;
                         break ;

        case OP_COLON  : CurrParse = &parsed[CurrParse[1]] ;
                         break ;

        default        : if (*CurrParse < 127 && isprint(*CurrParse))
                           SPRINTF(ErrBuf, "run: unknown operator '%c' (%d)",
                                   *CurrParse, *CurrParse) ;
                         else
                           SPRINTF(ErrBuf, "run: unknown operator %d",
                                   *CurrParse) ;
                         error(ERR_SNARK) ;
                         return ;
                      }
                  }
                ++ap ;
                ++rp ;
              }
          }

#if  SEQPAR
        if (m_myid == 0)
#endif  /* SEQPAR */
          disp_percentdone(y * 100 / (Ysize-1)) ;
      }
}


void
run()
{
#if  SEQPAR
  static int parked = 0 ;

  if (parked)
    m_rele_procs() ;
  m_fork(prun) ;
  m_kill_procs() ;     /* For some reason, parking doesn't work properly ...
  m_park_procs() ;
  parked = 1 ;         /* so it can be released above */
#else  /* ! SEQPAR */
  prun() ;
#endif  /* ! SEQPAR */

  disp_percentdone(100) ;  /* Can rely on '100' to terminate */
  CurNew->ncolors = col_used ;

  if (disp_active) drawimg(CurNew) ;

  SwapOldNew() ;
}


void
drawimg(old)
struct SRC *old ;
{
  int i, y ;
  unsigned char red[CMAPLEN], green[CMAPLEN], blue[CMAPLEN] ;
  struct SRC *new ;

  disp_imgstart(old->width, old->height, old->ncolors, old->Xalloc) ;
  if (old->ncolors == 3 && dtype == IS_COLOR)
    {
      if (scr_depth == 8) new = dither24to8(old) ;
      else                new = old ;
      for (y = 0; y < new->height; y++)
        disp_putline((pixel_t **) new->pix, y, new->width, new->ncolors) ;
      if (scr_depth == 8) ImgFree(new) ;
    }
  else if (dtype == IS_GRAY || dtype == IS_COLOR)
    {
      for (i = 0; i < CMAPLEN; i++) red[i] = green[i] = blue[i] = i ;
      disp_colormap(CMAPLEN, red, green, blue) ;
      for (y = 0; y < old->height; y++)
        disp_putline((pixel_t **) old->pix, y, old->width, old->ncolors) ;
#ifdef WANTED
      new = compress_gray(old) ;
      for (y = 0; y < new->height; y++)
        disp_putline((pixel_t **) new->pix, y, new->width, new->ncolors) ;
      ImgFree(new) ;
#endif /*WANTED*/
    }
  else
    {
      for (y = 0; y < old->height; y++)
        disp_putline((pixel_t **) old->pix, y, old->width, old->ncolors) ;
    }
  disp_imgend() ;
}


/*  Convert the three RGB lines into one NTSC luma line, for display
 *  on grayscale devices.  (Just pass pointer along if colors == 1)
 */

pixel_t*
ntsc_luma(lines, y, width)
pixel_t **lines ;
int y, width ;
{
  int off, x ;

  if (colors == 1) return(lines[0]) ;

  CLEARMEM((char *) ntsc, width * (sizeof(pixel_t))) ;
  off = y * width ;
  for (x = 0; x < width; x++)
    ntsc[x] = lines[0][off+x] * 0.299 +       /* R */
              lines[1][off+x] * 0.587 +       /* G */
              lines[2][off+x] * 0.114 ;       /* B */

  return(ntsc) ;
}
